# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for the grampg components: the PasswordGenerator class and each of
the Spec subclasses.
"""

import unittest

# Import internals for testing.
from grampg import grampg


def string_of_ints(string):
    """Whether `string` is a string of integers."""
    if not isinstance(string, str):
        return False
    for char in string:
        try:
            int(char)
        except TypeError:
            return False
    return True


def this_many_numbers(quantity, string):
    """Whether `string` is a string composed of `quantity` *number* chars."""
    if not string_of_ints(string) or len(string) != quantity:
        return False
    return True


def range_of_numbers(lbound, ubound, string):
    """Whether `string` is a string of a `quantity` *number* chars in range."""
    if not string_of_ints(string):
        return False
    if len(string) < lbound or len(string) > ubound:
        return False
    return True


class TestPasswordGenerator(unittest.TestCase):
    """
    Test creation and setup of the password generator.
    """

    def setUp(self):
        self.password = grampg.PasswordGenerator()

    def test_create_is_instance(self):
        """Creation always succeds."""
        self.assertIsInstance(self.password, grampg.PasswordGenerator)

    def test_create_has_sets_dict(self):
        """Expose a dictionary to easily handle character sets."""
        self.assertIsInstance(self.password.sets, dict)

    def test_of(self):
        """The initial generator instance always succeds."""
        gen = self.password.of()
        self.assertIsInstance(gen, grampg.PasswordGenerator)


class TestSpecsIndividually(unittest.TestCase):
    """
    Tests each spec individually, using the default character sets.
    """

    def setUp(self):
        # Extract the Generator instance from the builder.
        self.gen = grampg.PasswordGenerator().of().generator

    def test_generator_specs(self):
        """A generator exposes a `specs` dictionary."""
        self.assertIsInstance(self.gen.specs, dict)

    # Tests for the `exactly` spec method.

    def test_spec_exactly_instance(self):
        """Creation of spec."""
        self.gen.exactly(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertIsInstance(spec, grampg.CharacterExact)

    def test_spec_exactly_attrs(self):
        """`quantity` attribute is set."""
        self.gen.exactly(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertEqual(spec.low, 3)
        self.assertEqual(spec.high, 3)

    def test_spec_exactly_generate(self):
        """Generate correct number of characters."""
        self.gen.exactly(3, 'numbers')
        self.gen.done()
        res = self.gen.generate()
        self.assertTrue(this_many_numbers(3, res))

    def test_spec_exactly_generate_twice(self):
        """Calling generate twice results in correct characters."""
        self.gen.exactly(3, 'numbers')
        self.gen.done()
        res1 = self.gen.generate()
        res2 = self.gen.generate()
        self.assertTrue(res1, res2)

    # Test for the `between` spec method.

    def test_spec_between_instance(self):
        """Creation of "between" spec."""
        self.gen.between(3, 5, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertIsInstance(spec, grampg.CharacterRange)

    def test_spec_between_attrs(self):
        """`low` and `high` attributes are set in the spec."""
        self.gen.between(3, 5, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertEqual(spec.low, 3)
        self.assertEqual(spec.high, 5)

    def test_spec_between_generate(self):
        """Generate correct number of attributes."""
        self.gen.between(3, 5, 'numbers')
        self.gen.done()
        res = self.gen.generate()
        self.assertTrue(range_of_numbers(3, 5, res))

    # Test for the `at_least' spec method.

    def test_spec_at_least_instance(self):
        """Creation of the `between` spec."""
        self.gen.at_least(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertIsInstance(spec, grampg.CharacterRange)

    def test_spec_at_least_attrs(self):
        """When only lower bound is given the spec is not initialized."""
        self.gen.at_least(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertFalse(spec.initialized)

    def test_spec_at_least_generate(self):
        """Generate with no upper bound should fail."""
        self.gen.at_least(3, 'numbers')
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    def test_spec_at_least_generate_twice(self):
        """Calling generate twice results in two exceptions."""
        self.gen.at_least(3, 'numbers')
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    # Test for the `at_most` spec method.

    def test_spec_at_most_instance(self):
        """Creation of the `between` spec."""
        self.gen.at_most(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertIsInstance(spec, grampg.CharacterRange)

    def test_spec_at_most_attrs(self):
        """`high` attribute is set in the spec (disregard `low`)."""
        self.gen.at_most(3, 'numbers')
        spec = self.gen.specs['numbers']
        self.assertEqual(spec.high, 3)

    def test_spec_at_most_generate(self):
        """Generate with no lower bound should generate an empty string."""
        self.gen.at_most(3, 'numbers')
        self.gen.done()
        res = self.gen.generate()
        self.assertEqual(res, '')

    def test_spec_at_most_generate_twice(self):
        """Calling generate twice results in two empty strings."""
        self.gen.at_most(3, 'numbers')
        self.gen.done()
        res1 = self.gen.generate()
        res2 = self.gen.generate()
        self.assertEqual(res1, res2)

    # Tests for the `length` spec method.

    def test_length_type(self):
        """Length must be an int."""
        self.assertRaises(ValueError, lambda: self.gen.length('lenght'))

    def test_length_number(self):
        """Lenght must be a natural number."""
        self.assertRaises(ValueError, lambda: self.gen.length(0))
        self.assertRaises(ValueError, lambda: self.gen.length(-2))

    def test_length_empty(self):
        """Specifying only the length generates an empty string."""
        self.gen.length(2)
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    # Tests for `beginning_with` spec method.

    def test_beginning_with_empty(self):
        """Specifying only the beginning is invalid."""
        self.gen.beginning_with('numbers')
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    def test_beginning_with_nonregistered_set(self):
        """Character set must be registered."""
        self.assertRaises(grampg.PasswordSpecsError,
                          lambda: self.gen.beginning_with('nothing'))

    # Tests for `ending_with` spec method.

    def test_ending_with_single(self):
        """Specifying only ending_with is invalid."""
        self.gen.ending_with('numbers')
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    def test_ending_with_nonregistered_set(self):
        """Character set must be registered."""
        self.assertRaises(grampg.PasswordSpecsError,
                          lambda: self.gen.ending_with('nothing'))

    # Tests for `done` spec method.

    def test_just_done(self):
        """No char sets associated raises validation error."""
        self.assertRaises(grampg.PasswordSpecsError, self.gen.done)

    def test_done_idempotency(self):
        """Calling done after done in a valid generator has no effect."""
        self.gen.exactly(3, 'numbers')
        self.gen.done()
        self.gen.done()
        self.gen.done()

    def test_done_freezes_generator(self):
        """No more specs allowed after done is called."""
        self.gen.exactly(3, 'numbers')
        self.gen.done()
        self.assertRaises(grampg.PasswordSpecsError,
                          lambda: self.gen.length(3))


if __name__ == '__main__':
    unittest.main()



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Demo program for the grampg library.

The many_passwords program generates two sets of passwords (the total number of
which is parameterized), each according to "reasonable" specs.
"""

import grampg


def many_passwords(quantity):
    password = grampg.PasswordGenerator()

    print ("Passwords of 3 lower letters and 3"
           " numbers beginning with a lower letter.")
    generator = (password.of()
                 .exactly(3, 'lower_letters')
                 .exactly(3, 'numbers')
                 .beginning_with('lower_letters')
                 .done())
    for i in xrange(quantity):
        print "\t", generator.generate()
    print ""

    print ("Passwords of between 3 and 7 lower letters and 5"
           " upper letters, with a fixed length of 10.")
    generator = (password.of()
                 .between(3, 7, 'lower_letters')
                 .exactly(5, 'upper_letters')
                 .length(10)
                 .done())
    for i in xrange(quantity):
        print "\t", generator.generate()
    print ""


if __name__ == '__main__':
    many_passwords(5)



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Demo programs for the grampg library:

* many_passwords.py : generate two sets of passwords according to schemas which
  are printed to stdout before each set. The two schemas are reasonable.
"""



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Usecase tests, part of the grampg test suite.
"""

import unittest

# Import internals for testing.
from grampg import grampg


def how_many(charset, string, pgenerator):
    """Return how many of `kind` are in `string`."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    counter = 0
    for char in string:
        if char in pgenerator.sets[charset]:
            counter += 1
    return counter


def between(lbound, rbound, charset, string, pgenerator):
    """If there's `charset` between [`lbound`, `rbound`] chars in `string`."""
    num = how_many(charset, string, pgenerator)
    return num >= lbound and num <= rbound


def begins_with(charset, string, pgenerator):
    """Whether `string` begins with a `charset` character."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    if not string:
        return False
    return string[0] in pgenerator.sets[charset]


def ends_with(charset, string, pgenerator):
    """Whether `string` ends with a `charset` character."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    if not string:
        return False
    return string[-1:] in pgenerator.sets[charset]


class TestUseCase(unittest.TestCase):
    """
    Test for several schemes, both valid and faulty.
    """

    def setUp(self):
        self.lower_letters = 'lower_letters'
        self.numbers = 'numbers'
        self.upper_letters = 'upper_letters'

        self.generator = grampg.PasswordGenerator().of()

    def test_five_lower_letters(self):
        """Five lower letters."""
        res = self.generator.exactly(5, self.lower_letters).done().generate()
        self.assertEqual(5, len(res))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))

    def test_five_lower_letters_and_five_numbers(self):
        """Five lower letters and five numbers."""
        res = (self.generator.exactly(5, self.lower_letters)
                             .exactly(5, self.numbers).done().generate())
        self.assertEqual(10, len(res))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))
        self.assertEqual(5, how_many(self.numbers, res, self.generator))

    def test_five_lower_letters_and_five_numbers_bad_length(self):
        """Five lower letters and five numbers, with a bad length."""
        gen = (self.generator.exactly(5, self.lower_letters)
                             .exactly(5, self.numbers)
                             .length(8))
        self.assertRaises(grampg.PasswordSpecsError, gen.done)

    def test_five_lower_letters_and_two_to_five_numbers(self):
        """Five lower letters and two to five numbers."""
        res = (self.generator.exactly(5, self.lower_letters)
                             .between(2, 5, self.numbers).done().generate())
        self.assertTrue(7 <= len(res) and len(res) <= 10)
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))
        self.assertTrue(between(2, 5, self.numbers, res, self.generator))

    def test_five_lower_letters_and_two_to_five_numbers_with_length(self):
        """Five lower letters and two to five numbers with length in range."""
        res = (self.generator.exactly(5, self.lower_letters)
                             .between(2, 5, self.numbers)
                             .length(8).done().generate())
        self.assertEqual(8, len(res))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))
        self.assertEqual(3, how_many(self.numbers, res, self.generator))

    def test_five_lower_letters_two_to_five_num_with_length_and_begin(self):
        """Five lower letters and two to five numbers, beginning with a lower
        letter with length in range."""
        res = (self.generator.exactly(5, self.lower_letters)
                             .between(2, 5, self.numbers)
                             .beginning_with(self.lower_letters)
                             .length(8).done().generate())
        self.assertEqual(8, len(res))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))
        self.assertEqual(3, how_many(self.numbers, res, self.generator))
        self.assertTrue(begins_with(self.lower_letters, res, self.generator))

    def test_five_lower_letters_and_two_to_five_num_with_length_and_ends(self):
        """Five lower letters and two to five numbers, ending with a lower
        letter with length in range."""
        res = (self.generator.exactly(5, self.lower_letters)
                             .between(2, 5, self.numbers)
                             .ending_with(self.lower_letters)
                             .length(8).done().generate())
        self.assertEqual(8, len(res))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))
        self.assertEqual(3, how_many(self.numbers, res, self.generator))
        self.assertTrue(ends_with(self.lower_letters, res, self.generator))

    def test_five_lower_letters_bad_length(self):
        """Five lower letter, with bad length."""
        gen = self.generator.exactly(5, self.lower_letters).length(8)
        self.assertRaises(grampg.PasswordSpecsError, gen.done)

    def test_two_to_five_lower_letters_bad_length(self):
        """Two to five lower letter, with bad length."""
        gen = self.generator.between(2, 5, self.lower_letters).length(8)
        self.assertRaises(grampg.PasswordSpecsError, gen.done)

    def test_five_numbers_and_at_least_eight_lower_letters(self):
        """Five numbers and at least eight lower letters... but how long."""
        gen = (self.generator.exactly(5, self.numbers)
                             .at_least(8, self.lower_letters))
        # No way of saying how many characters in total.
        self.assertRaises(grampg.PasswordSpecsError, gen.done)

    def test_five_numbers_and_at_least_eight_lower_letters_with_length(self):
        """Five numbers and at least eight lower letters, with length."""
        res = (self.generator.exactly(5, self.numbers)
                             .at_least(8, self.lower_letters)
                             .length(15)
                             .done()
                             .generate())
        self.assertEqual(15, len(res))
        self.assertEqual(5, how_many(self.numbers, res, self.generator))
        self.assertEqual(10, how_many(self.lower_letters, res, self.generator))

    def test_five_numbers_and_at_most_eight_lower_letters(self):
        """Five numbers and at most eight lower letters."""
        res = (self.generator.exactly(5, self.numbers)
                             .at_most(8, self.lower_letters)
                             .done()
                             .generate())
        # Minimum password length.
        self.assertEqual(5, len(res))
        self.assertEqual(5, how_many(self.numbers, res, self.generator))
        self.assertEqual(0, how_many(self.lower_letters, res, self.generator))

    def test_five_num_and_at_most_eight_lower_letters_with_good_length(self):
        """Five numbers and at most eight lower letters, with length."""
        res = (self.generator.exactly(5, self.numbers)
                             .at_most(8, self.lower_letters)
                             .length(10)
                             .done()
                             .generate())
        # Minimum password length.
        self.assertEqual(10, len(res))
        self.assertEqual(5, how_many(self.numbers, res, self.generator))
        self.assertEqual(5, how_many(self.lower_letters, res, self.generator))

    def test_five_num_and_at_most_eight_lower_letters_with_bad_length(self):
        """Five numbers and at most eight lower letters, with bad length."""
        gen = (self.generator.exactly(5, self.numbers)
                             .at_most(8, self.lower_letters)
                             .length(14))
        # Not enough characters.
        self.assertRaises(grampg.PasswordSpecsError, gen.done)


if __name__ == '__main__':
    unittest.main()



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Regression tests, from debuggin the grampg to its test suite.
"""

import unittest

# Import internals for testing.
from grampg import grampg


def how_many(charset, string, pgenerator):
    """Return how many of `charset` are in `string`."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    counter = 0
    for char in string:
        if char in pgenerator.sets[charset]:
            counter += 1
    return counter


def between(lbound, rbound, charset, string, pgenerator):
    """If there's `charset` between [`lbound`, `rbound`] chars in `string`."""
    num = how_many(charset, string, pgenerator)
    return num >= lbound and num <= rbound


def begins_with(charset, string, pgenerator):
    """Whether `string` begins with a `charset` character."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    if not string:
        return False
    return string[0] in pgenerator.sets[charset]


def ends_with(charset, string, pgenerator):
    """Whether `string` ends with a `charset` character."""
    if charset not in pgenerator.sets.keys():
        raise ValueError('Character set %s is not known.' % charset)
    if not string:
        return False
    return string[-1:] in pgenerator.sets[charset]


class TestUseCase(unittest.TestCase):
    """
    Test for several schemes, both valid and faulty.
    """

    def setUp(self):
        self.numbers = 'numbers'
        self.letters = 'letters'

        self.generator = grampg.PasswordGenerator().of()

    def test_at_least_at_most_with_length(self):
        """At least four numbers, at most 10 letters, with a total of 10."""
        gen = (self.generator.length(10)
                             .at_least(4, self.numbers)
                             .at_most(10, self.letters)
                             .done())
        pass1, pass2, pass3 = gen.generate(), gen.generate(), gen.generate()

        # Test the length.
        self.assertEqual(10, len(pass1))
        self.assertEqual(10, len(pass2))
        self.assertEqual(10, len(pass3))
        # Test the numbers.
        numbers_in_range = lambda p: between(4, 10, self.numbers, p, gen)
        self.assertTrue(numbers_in_range(pass1))
        self.assertTrue(numbers_in_range(pass2))
        self.assertTrue(numbers_in_range(pass3))
        # Test the letters.
        letters_in_range = lambda p: between(0, 10, self.letters, p, gen)
        self.assertTrue(letters_in_range(pass1))
        self.assertTrue(letters_in_range(pass2))
        self.assertTrue(letters_in_range(pass3))


if __name__ == '__main__':
    unittest.main()



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
Defines the exception hierarchy, helper funtions and the main classes.

The exception hierarchy is based on :exc:`PasswordSpecsError`, extended by
:exc:`PasswordSpecsNonValidatedError` and :exc:`PasswordGeneratorIsDone` to
signal improper usage.

For details about the password generation algorithm read the grampg/docs.

Use the :class:`PasswordGenerator` as interface.
"""

import random


class PasswordSpecsError(Exception):
    """
    Root of grampg exceptions.

    Itself used to signal errors during specification or validation of a
    generator.
    """
    pass


class PasswordSpecsNonValidatedError(PasswordSpecsError):
    """
    Raised when :meth:`~Generator.generate` is called on a generator before a
    it is *done*.
    """
    pass


class PasswordGeneratorIsDone(PasswordSpecsError):
    """
    Raised when a new specification is attempted on a *done* generator.
    """
    pass


###############################################################################


def infinite_iter(character_set):
    """Transform a character set into a generator-iterator.

    Each value yielded is a :func:`random.choice` on the `character_set`."""
    if not character_set:
        return
    # yield forever.
    while True:
        yield random.choice(character_set)


def choice_of_spec(specs):
    """Return a random spec element from the sequence of `specs`.

    The probability of picking a particular spec is directly dependant on the
    `len` of its associated character set (expected at spec.set).

    Raises :exc:`~exception.ValueError` if `specs` is an empty sequence.
    """
    if not specs:
        raise ValueError('specs must not be empty.')
    sample = sum(len(spec.set) for spec in specs)
    choice = random.randint(1, sample)
    for spec in specs:
        spec_len = len(spec.set)
        if choice <= spec_len:
            return spec
        choice -= spec_len


###############################################################################


class Spec(object):
    """
    An iterator-like object which choses randomly from the character set.

    Not meant to be used, but subclassed. This class defines the internal state
    of *specs*, and the API each should provide.
    """

    def __init__(self, character_set, low=0, high=0):
        """Prepare the iteration.

        The default values of `low` and `high` are so that these can be
        accumulated over multiple specs, to produce *minimum* and *maximum*
        length, if desired.

        Positional arguments:
        character_set -- the character set to chose from.
        low -- the lower bound of the spec (default 0).
        high -- the upper bound of the spec (default 0).
        """
        self.low = low
        self.high = high
        self.set = character_set
        # Initialize the character generator-iterator.
        self.chars = infinite_iter(character_set)
        # Characters yielded so far.
        self.yielded = 0
        # Syntax sugar.
        self.another_character = self.next

    def in_range(self):
        """Whether the amount of chars yielded so far satisfy this spec."""
        return self.low <= self.yielded and self.yielded <= self.high

    def reset(self):
        """Reset internal state, the count of yielded characters."""
        self.yielded = 0

    def next(self):
        """Return next randomly chosen char in the set."""
        if self.yielded == self.high:
            raise StopIteration()
        self.yielded += 1
        return self.chars.next()

    def __iter__(self):
        return self

    def fix(self, full_length=None):
        """Fix the range of the spec."""
        raise NotImplementedError()


class CharacterExact(Spec):
    """
    Specify an exact number of characters for a particular set.
    """

    def __init__(self, character_set, quantity):
        """Initialize the spec.

        Positional arguments:
        character_set -- the character set to chose characters from.
        quantity -- the number of characters to chose.
        """
        # self.low and self.high are both set to quantity.
        super(CharacterExact, self).__init__(character_set, quantity, quantity)

    def fix(self, full_length=None):
        """Exact specs are always fixed."""
        pass


class CharacterRange(Spec):
    """
    Specify a numeric range for the number of characters for a character set.

    The range may be open at lower or upper bound, but not both. Use the
    :meth:`fix` operation to turn the range into a closed one, suitable to be
    used during generation. Notice that you can only fix the upper bound of a
    range if the total length of the password has been specified (via a call to
    :meth:`length`).
    """

    def __init__(self, character_set, low=None, high=None):
        if high:
            if low:
                super(CharacterRange, self).__init__(character_set, low, high)
            else:
                super(CharacterRange, self).__init__(character_set, high=high)
            self.initialized = True
        elif low:
            # An unbound range cannot be initialize. Should be fixed.
            # Temporary storage.
            self.low = low
            self.character_set = character_set
            self.initialized = False
        else:
            raise PasswordSpecsError("Either low or high bound"
                                     " must be specified.")

    def fix(self, full_length=None):
        """Force upper bound to have a meaningful value... or error out."""
        if not self.initialized:
            # Fix the upper bound.
            if full_length:
                super(CharacterRange, self).__init__(self.character_set,
                                                     self.low, full_length)
                self.initialized = True
            else:
                raise PasswordSpecsError("No upper bound nor"
                                         " length specified.")


###############################################################################


class Generator:
    """
    The generator object.

    A generator instance undergoes three phases during its existance: create it
    with the character sets to choose from, specify it by calling its methods
    finalizing in a call to :meth:`done`, and generate passwords
    with it by calling its :meth:`generate` method.

    Character sets should not be modified once the generator is
    instantiated. If other character sets are required, a new instance should
    be used.

    During the specification, repeated calls to the same method (consecutively
    or otherwise) overrides previous calls, so it is not an error to call them
    more than once. Specification is over after a call :meth:`~Generator.done`
    succeds. Once *done*, the generator cannot be further spec'ed, and only
    calls to :meth:`generate` are valid (although it is possible to call
    :meth:`done` over and over again, it does not have effect).

    Any attempt to add new specs to a *done* generator will raise
    :exc:`PasswordGeneratorIsDone`.

    .. note::

       Generator instances should be built by means of
       :class:`PasswordGenerator`, and only the :meth:`generate` method should
       ever be directly called on instances of this class.
    """

    def __init__(self, sets):
        """Initialize the generator with the character sets to use.

        Positional parameters:
        * the character sets to be used.
        """
        self.sets = sets
        self._length = None

        self.begins_with = None
        self.begins_with_char = ''
        self.ends_with = None
        self.ends_with_char = ''

        self._min_length = 0
        self._max_length = 0

        # When the generator is done, no more specs will be allowed.
        self._done = False

        # Collect the spacs as they come.
        self.specs = {}

    def length(self, length):
        """Set the length of generated passwords."""
        if self._done:
            raise PasswordGeneratorIsDone()

        if type(length) != int or length < 1:
            raise ValueError('Length must a natural number.')
        self._length = length

    def exactly(self, quantity, setname):
        """
        Specify an exact amount of characters for this set to be included in
        the generated passwords.
        """
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.specs[setname] = CharacterExact(self.sets[setname], quantity)

    def between(self, low, high, setname):
        """
        Specify an fixed range of characters for this set to be included in
        the generated passwords.
        """
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.specs[setname] = CharacterRange(self.sets[setname], low, high)

    def at_least(self, low, setname):
        """
        Specify the minimum amount of characters for this set to be included in
         the generated passwords.
         """
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.specs[setname] = CharacterRange(self.sets[setname], low=low)

    def at_most(self, high, setname):
        """
        Specify the maximum amount of characters for this set to be included in
        the generated passwords.
        """
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.specs[setname] = CharacterRange(self.sets[setname], high=high)

    def beginning_with(self, setname):
        """The password should begin with a character from this set."""
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.begins_with = setname

    def ending_with(self, setname):
        """The password should end with a character from this set."""
        if self._done:
            raise PasswordGeneratorIsDone()

        if setname not in self.sets:
            raise PasswordSpecsError('Unknown character set.')
        self.ends_with = setname

    def done(self):
        """Signals that the generator is ready to start producing passwords.

        Note that this call *freezes* or *marks* the generator object, which
        will not be able to receive further specs. A successful call to `done`
        must be made before :meth:`generate` can be called.

        This method is idempotent and irreversible.

        Raises :exc:`PasswordSpecsError` if validation of the specs fails.
        """
        if self._done:
            return

        if not self.specs:
            raise PasswordSpecsError("No character set spec is associated to"
                                     " this generator.")

        # Minimum length spec'ed, which is expected by generate().
        self._min_length = 0
        # Maximum length spec'ed, which is expected by generate().
        self._max_length = 0
        # Fix and accumulate minimum and maximum lengths
        for spec in self.specs.values():
            spec.fix(full_length=self._length)

            self._min_length += spec.low
            self._max_length += spec.high

        # Length validation of the specifications.
        if self._length and (self._length < self._min_length
                             or self._length > self._max_length):
            raise PasswordSpecsError('Length value specified is not'
                                     ' compatible with the other specs.')

        # Mark generator as *done*, will not receive further specs.
        self._done = True

    def generate(self):
        """Return one generated password based on the collected specs.

        Can be called any number of times, each yielding a new, independant
        password.

        Raises :exc:`PasswordSpecsNonValidatedError` if the generator is not
        *done* (the :meth:`done` method has not yet been called).
        """

        # Validate the specifications are done.
        if not self._done:
            raise PasswordSpecsNonValidatedError(
                "generate() called before done().")

        # Edge case: no character specs (only length, begin or end).
        if not self.specs:
            return ''

        # First, satisfy the minimum length specificated.

        # mins collects the minimum acceptable result from each spec.
        mins = []
        for spec_name, spec in self.specs.items():
            # If this happens to be the one to begin the result, take notice.
            if self.begins_with == spec_name:
                self.begins_with_char = spec.another_character()
            # Or in case it's the one to end it.
            if self.ends_with == spec_name:
                self.ends_with_char = spec.another_character()

            # Get a minimum amount of characters to satisfy this spec.
            chars = []
            while not spec.in_range():
                chars.append(spec.another_character())
            mins.append(chars)

        # Get all chars chosen so far, from all sets, into a bag.
        bag = []
        for chars in mins:
            bag.extend(chars)

        # The minimum length of all specs might not suit the specified length.

        if self._length and self._min_length < self._length:
            # To advance the length of the password, chose randomly
            # a specified character set and select another character.
            remaining_specs = [spec for spec in self.specs.values()
                               if spec.in_range()]
            min_length = self._min_length
            while min_length < self._length:
                chosen_spec = choice_of_spec(remaining_specs)
                # Check the character sets, so as to remove capped ones.
                try:
                    bag.append(chosen_spec.another_character())
                    min_length += 1
                except StopIteration:
                    # Spec has reached maximum number of tolerated chars.
                    remaining_specs.remove(chosen_spec)

        # Shuffle the bag to produce the output password.
        random.shuffle(bag)
        password = self.begins_with_char + ''.join(bag) + self.ends_with_char

        # Before terminating, reset the internal state of the specs.
        for spec in self.specs.values():
            spec.reset()

        return password


class PasswordGenerator:
    """
    Build the password generator.

    Provides a fluent interface to build :class:`Generator` instances, by means
    of method chaining.

    Exposes the character sets. Default character sets are provided for upper
    and lower case letters (`upper_letters` and `lower_letters`, respectively,
    all mashed up in `letters`) and `numbers`. A conjunction of the three is
    also provided, under the name `alphanumeric`.

    A character set can be registered by keying its name to a list of eligible
    characters in the :attr:`sets` attribute, or by extending the default
    character sets during instantiation.
    """

    _lower_letters = 'abcdefghijklmnopqrstuvwxyz'
    _upper_letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    _numbers = '1234567890'
    _letters = _lower_letters + _upper_letters
    _alpha = _letters + _numbers

    def __init__(self, from_sets={}):
        """Allow extra character sets to be registered.

        Keyword paramenters:
        from_sets: a dict of char set name to list of chars (default is {}).
        """
        self.sets = {
            'lower_letters': list(PasswordGenerator._lower_letters),
            'upper_letters': list(PasswordGenerator._upper_letters),
            'letters': list(PasswordGenerator._letters),
            'numbers': list(PasswordGenerator._numbers),
            'alphanumeric': list(PasswordGenerator._alpha),
            }
        self.sets.update(from_sets)
        self.generator = None

    def of(self):
        """Commence a method chain building a fresh generator instance.

        The generator instanciated by this call is new, but the character sets
        fed to it are always the same (the ones configured during
        :meth:`__init__`). If a different character set is desired, a new
        instance of :class:`PasswordGenerator` is neccessary.

        The generator will be finalized by a :meth:`done` call, and
        then used by calling :meth:`~Generator.generate` on it.
        """
        self.generator = Generator(self.sets)
        return self

    def length(self, length):
        """*Spec method*: adjust the total length of passwords to generate."""
        self.generator.length(length)
        return self

    def exactly(self, quantity, setname):
        """*Spec method*: require exactly this many characters from the set."""
        self.generator.exactly(quantity, setname)
        return self

    def between(self, low, high, setname):
        """
        *Spec method*: require no less than `low` but no more than `high`
        characters from that set. This spec defines a range of characters.
        """
        self.generator.between(low, high, setname)
        return self

    def at_least(self, low, setname):
        """
        *Spec method*: require no less than `low` but no more than `high`
        characters from that set. This spec defines a range of characters.
        """
        self.generator.at_least(low, setname)
        return self

    def at_most(self, high, setname):
        """
        *Spec method*: require no more than `high` characters from that
        set. This spec defines a range of characters.
        """
        self.generator.at_most(high, setname)
        return self

    def beginning_with(self, setname):
        """*Spec method*: passwords will start with a char from this set.

        Some other *spec method* must be called to define a number or range for
        that same set. Beginning with characters not specified is an error.
        """
        self.generator.beginning_with(setname)
        return self

    def ending_with(self, setname):
        """*Spec method*: passwords will end with a char from this set.

        Some other *spec method* must be called to define a number or range for
        that same set. Ending with characters not specified is an error.
        """
        self.generator.ending_with(setname)
        return self

    def done(self):
        """Finalize the generator and return it.

        The returned instance can receive calls to :meth:`~Generator.generate`,
        each of which will produce an independent password complying with the
        specs.
        """
        self.generator.done()
        return self.generator



from setuptools import setup

setup(
    name="grampg",
    version="0.1.1",
    packages=['grampg', 'grampg.tests', 'grampg.demo'],

    description="Simple and flexible password generation library.",
    long_description=open('README.txt').read(),

    license="GNU Affero General Public License (LICENSE.txt)",

    test_suite='tests',

    # Metadata for upload to PyPI.
    author="Elvio Toccalino",
    author_email="elvio.toccalino@gmail.com",
    keywords="grumpy admin password generator",
    url="https://bitbucket.org/etoccalino/grampg",
)



# Copywrite 2012 Elvio Toccalino

# This file is part of grampg.
#
# grampg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# grampg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with grampg.  If not, see <http://www.gnu.org/licenses/>.

"""
grampg package interface.

The `PasswordGenerator` class allows to build generators though a
fluent interface, by chaining methods to declare user specifications,
culminated to a call to `PasswordGenerator.done`. During this last
call, specs are validated, and a `PasswordSpecsError` is raised if
validation fails.
"""

from grampg import PasswordGenerator, PasswordSpecsError
from grampg import PasswordGeneratorIsDone
from grampg import PasswordSpecsNonValidatedError
